# Django
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# AA Example App
from corptools.models import CharacterAudit, CorporationAudit

from .. import providers
from .utils import NoTokens, get_corp_token, update_corp_audit

logger = get_extension_logger(__name__)


@update_corp_audit(update_field="tracking")
def update_character_logins_from_corp(corp_id, full_update=False):
    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("Updating Logins for: {}".format(
        audit_corp.corporation))

    req_scopes = [
        'esi-corporations.track_members.v1',
        'esi-characters.read_corporation_roles.v1'
    ]
    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        raise NoTokens("Character Tracking")
    tracking = providers.esi_openapi.client.Corporation.GetCorporationsCorporationIdMembertracking(
        corporation_id=corp_id,
        token=token
    ).results(
        force_refresh=full_update,
        store_cache=False
    )

    for c in tracking:
        try:
            ca = CharacterAudit.objects.get(
                character__character_id=c.character_id)
            ca.last_known_login = c.logon_date
            ca.last_known_logoff = c.logoff_date
            ca.save()
        except CharacterAudit.DoesNotExist:
            pass

    all_chars = CharacterAudit.objects.filter(
        character__corporation_id=corp_id)
    all_chars.update(last_update_login=timezone.now())

    return f"Finished Logins for: {audit_corp.corporation}"



# Third Party
from eve_sde.models import ItemType
from model_utils import Choices

# Django
from django.db import models

from .audits import CharacterAudit, EveLocation


class Clone(models.Model):
    character = models.OneToOneField(CharacterAudit, on_delete=models.CASCADE)

    last_clone_jump_date = models.DateTimeField(null=True, default=None)
    last_station_change_date = models.DateTimeField(null=True, default=None)
    location_id = models.BigIntegerField()
    location_name = models.ForeignKey(
        EveLocation,
        on_delete=models.SET_NULL,
        null=True,
        default=None
    )
    _type_enum = Choices('station', 'structure')
    location_type = models.CharField(max_length=9, choices=_type_enum)

# Clone Models


class JumpClone(models.Model):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    jump_clone_id = models.IntegerField(null=True, default=None)
    location_id = models.BigIntegerField(null=True, default=None)
    location_name = models.ForeignKey(
        EveLocation,
        on_delete=models.SET_NULL,
        null=True,
        default=None
    )
    _type_enum = Choices('station', 'structure')
    location_type = models.CharField(
        max_length=9,
        choices=_type_enum,
        null=True,
        default=None
    )
    name = models.CharField(max_length=255, null=True, default=None)

# Implant Model


class Implant(models.Model):
    clone = models.ForeignKey(JumpClone, on_delete=models.CASCADE)
    type_name = models.ForeignKey(
        ItemType,
        on_delete=models.SET_NULL,
        null=True,
        default=None
    )



# Third Party
from eve_sde.models import ItemType

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# AA Example App
from corptools.models import (
    CorporationAudit,
    CorporationIndustryJob,
)

from .. import providers
from .utils import NoTokens, get_corp_token, update_corp_audit

logger = get_extension_logger(__name__)


@update_corp_audit(update_field="industry_jobs")
def corp_update_industry_jobs(corp_id: int, force_refresh: bool = False) -> str:
    """
    https://developers.eveonline.com/api-explorer#/operations/GetCorporationsCorporationIdIndustryJobs
    """
    req_scopes = ['esi-industry.read_corporation_jobs.v1']

    req_roles = ['Factory_Manager']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        raise NoTokens("Indy")

    _corporation: CorporationAudit = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)

    industry_jobs = providers.esi_openapi.client.Industry.GetCorporationsCorporationIdIndustryJobs(
        corporation_id=_corporation.corporation.corporation_id,
        include_completed=True,
        token=token
    ).results(
        force_refresh=force_refresh,
        store_cache=False
    )

    existing_pks: set[int] = set(
        # only one we care about.
        CorporationIndustryJob.objects.filter(
            corporation=_corporation,
            job_id__in=[
                e.job_id for e in industry_jobs
            ]
        ).values_list("job_id", flat=True)
    )
    type_ids: set[int] = set()
    new_events = []
    for event in industry_jobs:
        type_ids.add(event.blueprint_type_id)
        if event.product_type_id:
            type_ids.add(event.product_type_id)

        if event.job_id in existing_pks:
            _m: CorporationIndustryJob = CorporationIndustryJob.objects.get(
                corporation=_corporation,
                job_id=event.job_id
            )
            _m.completed_character_id = event.completed_character_id
            _m.completed_date = event.completed_date
            _m.end_date = event.end_date
            _m.pause_date = event.pause_date
            _m.status = event.status
            _m.successful_runs = event.successful_runs
            _m.save()
            continue

        _e = CorporationIndustryJob(
            corporation=_corporation,
            activity_id=event.activity_id,
            blueprint_id=event.blueprint_id,
            blueprint_location_id=event.blueprint_location_id,
            blueprint_type_id=event.blueprint_type_id,
            blueprint_type_name_id=event.blueprint_type_id,
            completed_character_id=event.completed_character_id,
            completed_date=event.completed_date,
            cost=event.cost,
            duration=event.duration,
            end_date=event.end_date,
            facility_id=event.facility_id,
            installer_id=event.installer_id,
            job_id=event.job_id,
            licensed_runs=event.licensed_runs,
            location_id=event.location_id,
            output_location_id=event.output_location_id,
            pause_date=event.pause_date,
            probability=event.probability,
            product_type_id=event.product_type_id,
            product_type_name_id=event.product_type_id,
            runs=event.runs,
            start_date=event.start_date,
            status=event.status,
            successful_runs=event.successful_runs
        )

        new_events.append(_e)

    if len(new_events):
        CorporationIndustryJob.objects.bulk_create(
            new_events,
            ignore_conflicts=True
        )

    return f"Updated industry jobs for {_corporation}"



# Standard Library
from typing import Union

# Django
from django.utils import timezone

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenError
from esi.exceptions import HTTPError, HTTPNotModified
from esi.models import Token

# AA Example App
from corptools.models import CorporationAudit, EveName

from .. import providers

logger = get_extension_logger(__name__)


class NoTokens(Exception):
    pass


def update_corp_audit(update_field: str = ""):
    def decorator(function):
        def wrapper(*args, **kwargs):
            audit_corp = CorporationAudit.objects.get(
                corporation__corporation_id=args[0]
            )
            try:
                result = function(*args, **kwargs)
                try:
                    setattr(
                        audit_corp, f"last_update_{update_field}", timezone.now())
                    setattr(
                        audit_corp, f"last_change_{update_field}", timezone.now())
                except Exception:
                    pass
                audit_corp.save()
                return result
            except HTTPNotModified:
                logger.info(f"Not modified {function} {args} {kwargs}")
                setattr(audit_corp, update_field, timezone.now())
                audit_corp.save()
            except HTTPError as e:
                logger.error(f"HTTPError {function} {args} {kwargs} {e}")
            except NoTokens as e:
                logger.error(
                    f"No Token Found For {audit_corp.corporation.corporation_name} - {e}")

            return
        return wrapper
    return decorator


def get_corp_token(corp_id: int, scopes: list, req_roles: Union[list, None, bool]):
    """
    Helper method to get a token for a specific character from a specific corp with specific scopes, where
    a character has specific in game roles.
    :param corp_id: Corp to filter on.
    :param scopes: array of ESI scope strings required on the token
    :param req_roles: array of roles, one of which is required on the character.
    :return: :class:esi.models.Token or None
    """

    # always add roles scope as a requirement.
    if 'esi-characters.read_corporation_roles.v1' not in scopes:
        scopes.append("esi-characters.read_corporation_roles.v1")

    # Find all characters in the corporation known to auth.
    char_ids = EveCharacter.objects.filter(
        corporation_id=corp_id
    ).values('character_id')

    # find all tokens for the corp, with the scopes.
    tokens = Token.objects.filter(
        character_id__in=char_ids
    ).require_scopes(scopes)

    # loop to check the roles and break on first correct match
    for token in tokens:
        try:
            if req_roles:  # There are endpoints with no requirements
                roles = providers.esi_openapi.client.Character.GetCharactersCharacterIdRoles(
                    character_id=token.character_id,
                    token=token
                ).result(
                    use_etag=False
                )

                has_roles = False

                # do we have the roles.
                for role in roles.roles:
                    if role in req_roles:
                        has_roles = True
                        break

                if has_roles:
                    return token  # return the token
                else:
                    pass  # next! TODO should we flag this character?
            else:
                return token  # no roles check needed return the token
        except TokenError as e:
            #  I've had invalid tokens in auth that refresh but don't actually work
            logger.error(f"Token Error ID: {token.pk} ({e})")

    return None


def get_eve_ids(ids: list | set):
    return set(
        EveName.objects.filter(
            eve_id__in=list(ids)
        ).distinct().values_list(
            'eve_id',
            'name'
        )
    )



# Standard Library
from random import random

# Third Party
from celery import chain as Chain
from celery import shared_task

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from esi.exceptions import HTTPClientError, HTTPNotModified, HTTPServerError

from ...models import CorporationAudit
from .. import app_settings
from ..utils import esi_error_retry, no_fail_chain, rate_limited_task_no_fail
from .assets import corp_update_assets
from .characters import update_character_logins_from_corp
from .contracts import corp_contract_item_fetch, corp_contract_update
from .indy import corp_update_industry_jobs
from .structures import (
    corp_starbase_update,
    corp_structure_update,
    corp_update_pocos,
)
from .wallet import update_corp_wallet_divisions

TZ_STRING = "%Y-%m-%dT%H:%M:%SZ"


logger = get_extension_logger(__name__)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_wallet"
)
@no_fail_chain
@esi_error_retry
def update_corp_wallet(self, corp_id, force_refresh=False, chain=[]):
    return update_corp_wallet_divisions(corp_id, full_update=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_structures"
)
@no_fail_chain
@esi_error_retry
def update_corp_structures(self, corp_id, force_refresh=False, chain=[]):
    return corp_structure_update(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_assets"
)
@esi_error_retry
def update_corp_assets(self, corp_id, force_refresh=False, chain=[]):
    return corp_update_assets(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_pocos"
)
@no_fail_chain
@esi_error_retry
def update_corp_pocos(self, corp_id, force_refresh=False, chain=[]):
    return corp_update_pocos(corp_id, full_update=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_logins"
)
@no_fail_chain
@esi_error_retry
def update_corp_logins(self, corp_id, force_refresh=False, chain=[]):
    return update_character_logins_from_corp(corp_id, full_update=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_starbases"
)
@no_fail_chain
@esi_error_retry
def update_corp_starbases(self, corp_id, force_refresh=False, chain=[]):
    return corp_starbase_update(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_contract_items"
)
@rate_limited_task_no_fail("600/15m")
@esi_error_retry
def update_corporate_contract_items(self, corp_id, contract_id, force_refresh=False):
    try:
        return corp_contract_item_fetch(corp_id, contract_id, force_refresh=force_refresh)
    except HTTPNotModified:
        pass
    except (HTTPClientError, HTTPServerError) as e:
        if e.status_code == 404:
            _msg = f"CT: Contract items {str(corp_id)} ({str(contract_id)}) NOT FOUND ERROR"
            logger.warning(
                f"CT: Contract items {str(corp_id)} ({str(contract_id)}) NOT FOUND ERROR"
            )
            return f"CT: Contract items {str(corp_id)} ({str(contract_id)}) NOT FOUND ERROR"
        else:
            self.retry()


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_contracts"
)
@no_fail_chain
@esi_error_retry
def update_corp_contracts(self, corp_id, force_refresh=False, chain=[]):
    ret = corp_contract_update(
        corp_id,
        force_refresh=force_refresh
    )

    if ret:
        _chain = []
        for id in ret[1]:
            _chain.append(
                update_corporate_contract_items.si(
                    corp_id,
                    id
                )
            )
        Chain(_chain).apply_async(priority=8)

    return "Completed Que of contract items for: %s" % str(corp_id)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_industry_jobs"
)
@no_fail_chain
@esi_error_retry
def update_corp_industry_jobs(self, corp_id: int, force_refresh: bool = False, chain=[]) -> str:
    return corp_update_industry_jobs(corp_id, force_refresh=force_refresh)


@shared_task(
    name="corptools.tasks.update_corp"
)
def update_corp(corp_id, force_refresh=False):
    corp = CorporationAudit.objects.get(corporation__corporation_id=corp_id)
    logger.info(f"Starting Updates for {corp.corporation.corporation_name}")

    que = []
    que.append(update_corp_wallet.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_structures.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_starbases.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_assets.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_pocos.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_contracts.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_logins.si(corp_id, force_refresh=force_refresh))
    que.append(
        update_corp_industry_jobs.si(
            corp_id,
            force_refresh=force_refresh
        )
    )

    Chain(que).apply_async(priority=6)


@shared_task(
    name="corptools.tasks.update_all_corps"
)
def update_all_corps(force_refresh=False):
    corps = CorporationAudit.objects.all().select_related('corporation')
    for corp in corps:
        countdown = 1  # if force_refresh else random()*app_settings.CT_TASK_SPREAD_DELAY*2
        update_corp.apply_async(
            args=[corp.corporation.corporation_id],
            kwargs={
                "force_refresh": force_refresh
            },
            countdown=countdown
        )



# Third Party
from eve_sde.models import ItemType

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from esi.exceptions import HTTPNotModified

# AA Example App
from corptools.models import (
    CorporateContract,
    CorporateContractItem,
    CorporationAudit,
    EveName,
)

from .. import providers
from .utils import NoTokens, get_corp_token, update_corp_audit

logger = get_extension_logger(__name__)


@update_corp_audit(update_field="contracts")
def corp_contract_update(corp_id, force_refresh=False):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id
    )

    logger.debug(
        "updating corporate contracts for: %s" % str(
            _corporation.corporation.corporation_name)
    )

    req_scopes = [
        'esi-contracts.read_corporation_contracts.v1',
        'esi-characters.read_corporation_roles.v1'
    ]

    token = get_corp_token(corp_id, req_scopes, False)

    new_contract_ids = []

    if not token:
        raise NoTokens("Contracts")

    contracts = providers.esi_openapi.client.Contracts.GetCorporationsCorporationIdContracts(
        corporation_id=corp_id,
        token=token
    ).results(
        force_refresh=force_refresh,
        store_cache=False
    )

    contract_models_new = []
    contract_models_old = []
    eve_names = set()
    contract_ids = list(
        CorporateContract.objects.filter(
            corporation=_corporation
        ).values_list(
            "contract_id",
            flat=True
        )
    )
    for c in contracts:  # update labels
        _contract_item = CorporateContract.from_esi_model(_corporation, c)
        eve_names.add(c.assignee_id)
        eve_names.add(c.issuer_corporation_id)
        eve_names.add(c.issuer_id)
        eve_names.add(c.acceptor_id)

        if c.contract_id in contract_ids:
            contract_models_old.append(_contract_item)
        else:
            contract_models_new.append(_contract_item)

    EveName.objects.create_bulk_from_esi(list(eve_names))

    if len(contract_models_new) > 0:
        CorporateContract.objects.bulk_create(
            contract_models_new, batch_size=1000, ignore_conflicts=True)

    if len(contract_models_old) > 0:
        CorporateContract.objects.bulk_update(
            contract_models_old,
            batch_size=1000,
            fields=[
                'date_accepted',
                'date_completed',
                'acceptor_id',
                'date_expired',
                'status',
                'assignee_name',
                'acceptor_name',
                'issuer_corporation_name',
                'issuer_name'
            ]
        )

    new_contract_ids = [c.contract_id for c in contract_models_new]
    if force_refresh:
        new_contract_ids += [c.contract_id for c in contract_models_old]

    return "CT: Completed corporate contracts for: %s" % str(_corporation.corporation.corporation_name), new_contract_ids


def corp_contract_item_fetch(corp_id, contract_id, force_refresh=False):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id
    )

    logger.debug(
        "updating contract items for: %s (%s)" % (
            str(_corporation.corporation.corporation_name),
            str(contract_id)
        )
    )

    contract = CorporateContract.objects.get(
        corporation=_corporation,
        contract_id=contract_id)

    if contract.status != "deleted":
        req_scopes = [
            'esi-contracts.read_corporation_contracts.v1',
            'esi-characters.read_corporation_roles.v1'
        ]

        token = get_corp_token(corp_id, req_scopes, False)

        if not token:
            return
        try:
            contracts = providers.esi_openapi.client.Contracts.GetCorporationsCorporationIdContractsContractIdItems(
                corporation_id=corp_id,
                contract_id=contract_id,
                token=token
            ).results(
                force_refresh=force_refresh,
                store_cache=False
            )

            contract_models_new = []
            _types = set()
            for c in contracts:  # update labels
                _contract_item = CorporateContractItem.from_esi_model(
                    contract, c)
                _types.add(c.type_id)
                contract_models_new.append(_contract_item)
            CorporateContractItem.objects.bulk_create(
                contract_models_new, batch_size=1000, ignore_conflicts=True)
        except HTTPNotModified:
            logger.info(
                f"CT: No New Corporate Contract items {str(_corporation.corporation.corporation_name)} ({str(contract_id)})")

    return f"CT: Completed Corporate Contract items {str(_corporation.corporation.corporation_name)} ({str(contract_id)})"



# Standard Library
import re

# Third Party
import networkx as nx

# Django
from django.utils import timezone

# Alliance Auth
from esi.openapi_clients import ESIClientProvider as ESIOpenApiProvider

from . import __appname__, __url__, __version__, app_settings
from .task_helpers.skill_helpers import SkillListCache


class EveRouter():
    def __init__(self):
        self.G = nx.Graph()
        self.last_update = None
        self.bridges = {}

    def bulid_graph(self):
        self.G.clear()
        # logger.debug("Graph cleared.")
        from .models import MapJumpBridge, MapSystem, MapSystemGate

        systems = MapSystem.objects.values_list('system_id', flat=True)
        gates = MapSystemGate.objects.values_list(
            'from_solar_system_id', 'to_solar_system_id')
        bridges_query = MapJumpBridge.objects.values_list(
            'from_solar_system_id', 'to_solar_system_id', 'structure_id')
        bridges = []
        bridge_ids = {}
        for b in bridges_query:
            bridge_ids[b[0]] = b[2]
            bridges.append([b[0], b[1]])
        self.G.add_nodes_from(systems)
        self.G.add_edges_from(gates, type="gate")
        self.G.add_edges_from(bridges, type="bridge")
        self.bridges = bridge_ids
        self.last_update = timezone.now()

    def route(self, source_id, destination_id):
        from .models import MapJumpBridge, MapSystem

        # logger.debug("----------")
        # logger.debug(f"Source: {source_id}")
        # logger.debug(f"Destination: {destination_id}")

        if not self.last_update:
            self.bulid_graph()
        else:
            try:
                last_update = MapJumpBridge.objects.latest("updated").updated
                if last_update > self.last_update:
                    self.bulid_graph()
            except MapJumpBridge.DoesNotExist:
                pass

        path = nx.shortest_path(self.G, source_id, destination_id)
        path_length = len(path) - 1
        systems = MapSystem.objects.filter(system_id__in=path)
        system_map = {}

        for a in systems:
            system_map[a.system_id] = a.name

        output = {}
        for i, p in enumerate(path):
            output[i] = system_map[p]

        dotlan_path = output[0].replace(" ", "_")
        message_path = "`" + output[0].replace(" ", "_") + "`"
        esi_points = []
        for i in range(len(path) - 1):
            if self.G.get_edge_data(path[i], path[i + 1])['type'] == 'bridge':
                message_path += "  >B>  `" + output[i + 1] + "`"
                bridged_path = '::' + output[i + 1].replace(" ", "_")
                dotlan_path += bridged_path
                esi_points.append(self.bridges[path[i]])
            else:
                message_path += "  >G>  `" + output[i + 1] + "`"
                gated_path = ':' + output[i + 1].replace(" ", "_")
                dotlan_path += gated_path
        dotlan_path = re.sub(r'(?<!:)(:[^:\s]+)(?=:)(?!::)', '', dotlan_path)
        esi_points.append(path[len(path) - 1])
        result = {'path': output, 'esi': esi_points, 'path_message': message_path,
                  'dotlan': dotlan_path, 'length': path_length}

        return result


esi = None
routes = EveRouter()
skills = SkillListCache()

compat = "2025-08-26" if not app_settings.CT_COMPAT_DATE_OVERRIDE else app_settings.CT_COMPAT_DATE_OVERRIDE


class OpenAPI(ESIOpenApiProvider):
    def char_location_op(self, character_id, token):
        return self.client.Location.GetCharactersCharacterIdLocation(
            character_id=character_id,
            token=token
        )

    def char_location(self, character_id, token, force_refresh=False):
        return self.char_location_op(
            character_id,
            token
        ).result(
            force_refresh=force_refresh,
            use_etag=False
        )

    @staticmethod
    def chunk_ids(lo, n=750):
        for i in range(0, len(lo), n):
            yield lo[i:i + n]


esi_openapi = OpenAPI(
    compatibility_date=compat,
    ua_appname=__appname__,
    ua_url=__url__,
    ua_version=__version__,
    operations=[
        # Char Audits
        "GetCharactersCharacterIdCorporationhistory",
        "GetCharactersCharacterIdRoles",
        "GetCharactersCharacterIdTitles",
        # Character Asset
        "GetCharactersCharacterIdAssets",
        "PostCharactersCharacterIdAssetsNames",
        "PostCharactersCharacterIdAssetsLocations",
        # Character Location
        "GetCharactersCharacterIdShip",
        "GetCharactersCharacterIdLocation",
        "GetCharactersCharacterIdOnline",
        # Character Interactions
        "GetCharactersCharacterIdNotifications",
        "GetCharactersCharacterIdLoyaltyPoints",
        "GetCharactersCharacterIdContacts",
        "GetCharactersCharacterIdContactsLabels",
        # Corporate Assets
        "GetCorporationsCorporationIdAssets",
        "PostCorporationsCorporationIdAssetsNames",
        "PostCorporationsCorporationIdAssetsLocations",
        # Character Skills
        "GetCharactersCharacterIdSkills",
        "GetCharactersCharacterIdSkillqueue",
        # Character Indy
        "GetCharactersCharacterIdIndustryJobs",
        "GetCharactersCharacterIdMining",
        # Character Clones
        "GetCharactersCharacterIdClones",
        "GetCharactersCharacterIdImplants",
        # Character Contracts
        "GetCharactersCharacterIdContracts",
        "GetCharactersCharacterIdContractsContractIdItems",
        # Character Wallets
        "GetCharactersCharacterIdWallet",
        "GetCharactersCharacterIdWalletJournal",
        "GetCharactersCharacterIdWalletTransactions",
        "GetCharactersCharacterIdOrdersHistory",
        "GetCharactersCharacterIdOrders",
        # Corp Data
        "GetCorporationsCorporationIdDivisions",
        # Universe
        "GetUniverseTypesTypeId",
        "GetUniverseGroupsGroupId",
        "GetUniverseCategoriesCategoryId",
        "PostUniverseNames",
        "PostUniverseIds",
        # Corp Wallets
        "GetCorporationsCorporationIdWalletsDivisionJournal",
        "GetCorporationsCorporationIdWalletsDivisionTransactions",
        "GetCorporationsCorporationIdDivisions",
        "GetCorporationsCorporationIdWallets",
        "GetCorporationsCorporationIdContracts",
        "GetCorporationsCorporationIdContractsContractIdItems",
        # Corp Assets
        "GetCorporationsCorporationIdStructures",
        "GetCorporationsCorporationIdStarbases",
        "GetCorporationsCorporationIdStarbasesStarbaseId",
        "GetCorporationsCorporationIdCustomsOffices",
        # Corp Indy
        "GetCorporationsCorporationIdIndustryJobs",
        # Corp Members
        "GetCorporationsCorporationIdMembertracking",
        # Mail endpoitns
        "GetCharactersCharacterIdMailLabels",
        "GetCharactersCharacterIdMailMailId",
        "GetCharactersCharacterIdMail",
        # Misc
        "GetUniverseStationsStationId",
        "GetUniverseStructuresStructureId",
        "GetUniverseSystemsSystemId",
        "GetStatus",
        "GetUniverseCategories",
        # moons
        "GetCorporationCorporationIdMiningObservers",
        "GetCorporationCorporationIdMiningObserversObserverId",
    ]
)



# Standard Library
import re
from typing import Union

# Third Party
from aiopenapi3 import HTTPError
from celery import chain, shared_task
from eve_sde.models import ItemType

# Django
from django.db.models import F, Q
from django.db.models.aggregates import Sum

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

# AA Example App
from corptools.models import (
    AssetCoordiante,
    BridgeOzoneLevel,
    CorpAsset,
    CorporationAudit,
    EveLocation,
    Structure,
)
from corptools.task_helpers.update_tasks import fetch_location_name

from .. import providers
from ..utils import chunks
from .utils import NoTokens, get_corp_token, update_corp_audit

logger = get_extension_logger(__name__)


@update_corp_audit(update_field="assets")
def corp_update_assets(corp_id, force_refresh: bool = False):
    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("Updating Assets for: {}".format(
        audit_corp.corporation))

    req_scopes = [
        'esi-assets.read_corporation_assets.v1',
        'esi-characters.read_corporation_roles.v1'
    ]

    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        raise NoTokens("Assets")

    failed_locations = []

    assets = providers.esi_openapi.client.Assets.GetCorporationsCorporationIdAssets(
        corporation_id=corp_id,
        token=token
    ).results(
        force_refresh=force_refresh,
        store_cache=False
    )

    location_names = list(
        EveLocation.objects.all().values_list('location_id', flat=True)
    )

    _current_type_ids = []
    item_ids = []
    items = []

    for item in assets:
        if item.type_id not in _current_type_ids:
            _current_type_ids.append(item.type_id)
        item_ids.append(item.item_id)

        asset_item = CorpAsset.from_esi_model(audit_corp, item)

        if item.location_id not in location_names:
            try:
                if item.location_id not in failed_locations:
                    new_name = fetch_location_name(
                        item.location_id,
                        item.location_flag,
                        token.character_id
                    )
                    if new_name:
                        new_name.save()
                        location_names.append(
                            item.location_id
                        )
                        asset_item.location_name_id = item.location_id
                    else:
                        failed_locations.append(item.location_id)
            except Exception:
                pass  # TODO
        else:
            asset_item.location_name_id = item.location_id

        items.append(asset_item)

    delete_query = CorpAsset.objects.filter(
        corporation=audit_corp)  # Flush Assets
    if delete_query.exists():
        # with coords we need to care about the fkeys/signals
        delete_query.delete()

    CorpAsset.objects.bulk_create(items)
    try:
        corp_update_asset_names(corp_id)
    except Exception as e:
        logger.error(e)

    que = []
    que.append(fetch_coordiantes.si(corp_id, force_refresh=force_refresh))
    que.append(run_ozone_levels.si(corp_id))
    que.append(build_managed_asset_locations.si(
        corp_id, force_refresh=force_refresh))
    chain(que).apply_async(priority=7)

    return f"Finished assets for: {audit_corp.corporation}"


@shared_task(bind=True)
def build_managed_asset_locations(self, corp_id, force_refresh: bool = True):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("Updating Corporate Hangar Locations for: %s" %
                 (_corporation.corporation))

    req_scopes = ['esi-corporations.read_divisions.v1',
                  'esi-characters.read_corporation_roles.v1']
    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    divisions = {}
    if token:
        divs = providers.esi_openapi.client.Corporation.GetCorporationsCorporationIdDivisions(
            corporation_id=_corporation.corporation.corporation_id,
            token=token
        ).result(
            force_refresh=force_refresh,
            use_etag=False,
            store_cache=False
        )
        for d in divs.hangar:
            divisions[f"CorpSAG{d.division}"] = d.name

    existing_locations = set(list(EveLocation.objects.filter(
        managed=True, managed_corp=_corporation).values_list("location_id", flat=True)))
    current = []
    new = []
    updates = []
    # get office folders and name them
    folders = CorpAsset.objects.filter(
        corporation=_corporation,
        location_flag="OfficeFolder"
    ).select_related("location_name")

    names = {}

    grp_ids = [
        12,
        340,
        448,
        649,
    ]

    cat_ids = [
        6
    ]

    corp_hangars = {}
    hangar_ids = []
    for i in folders:
        system = None
        if i.location_name:
            names[i.item_id] = i.location_name.location_name
            system = i.location_name.system
        else:
            names[i.item_id] = i.location_id
        corp_hangars[i.item_id] = {}
        for j in range(1, 8):

            # Build the Managed Location for hangars
            id = -int(f"{i.item_id}{j}")
            sag = f"CorpSAG{j}"
            name = f"{names[i.item_id]} > {divisions.get(sag, sag)}"
            hangar_ids.append(id)

            # store for later?
            corp_hangars[i.item_id][sag] = (id, name)

            # update or create the location
            EveLocation.objects.update_or_create(
                location_id=id,
                defaults={
                    "location_name": name,
                    "managed": True,
                    "managed_corp": _corporation,
                    "system": system
                }
            )

            # update all the assets.
            CorpAsset.objects.filter(
                corporation=_corporation,
                location_id=i.item_id,
                location_flag=sag
            ).update(location_name_id=id)
    # get Containers in office folders and assign the names

    containers = CorpAsset.objects.filter(
        Q(type_name__group_id__in=grp_ids) | Q(
            type_name__group__category_id__in=cat_ids)
    ).filter(
        corporation=_corporation,
        singleton=True,
        location_id__in=folders.values_list(
            "item_id", flat=True),
    ).select_related("location_name", "type_name")

    req_scopes = ['esi-assets.read_corporation_assets.v1',
                  'esi-characters.read_corporation_roles.v1']

    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)
    if token:
        ids = list(set(list(containers.values_list("item_id", flat=True))))
        for c in chunks(ids, 900):
            items = providers.esi_openapi.client.Assets.PostCorporationsCorporationIdAssetsNames(
                body=c,
                corporation_id=_corporation.corporation.corporation_id,
                token=token
            ).result(
                force_refresh=force_refresh
            )
            for n in items:
                names[n.item_id] = n.name

    # each container gets a location name Station > Hangar > Type
    for c in containers:
        current.append(c.item_id)
        name = names.get(c.item_id, c.type_name.name)
        hangar = divisions.get(c.location_flag, c.location_flag)
        station = names.get(c.location_id, c.location_id)
        item_loc = f"{station} > {hangar} > {name}"
        loc = EveLocation(
            location_id=c.item_id,
            location_name=item_loc,
            managed=True,
            managed_corp=_corporation
        )
        if c.item_id in existing_locations:
            updates.append(loc)
        else:
            new.append(loc)

    if len(new):
        EveLocation.objects.bulk_create(new)

    if len(updates):
        EveLocation.objects.bulk_update(
            updates, ["location_name", "managed", "managed_corp"])

    EveLocation.objects.filter(managed=True, managed_corp=_corporation).exclude(
        location_id__in=current + hangar_ids).delete()

    CorpAsset.objects.filter(location_id__in=current).update(
        location_name_id=F("location_id"))


@shared_task(bind=True, base=QueueOnce)
def fetch_coordiantes(self, corp_id, force_refresh: bool = False):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id
    )

    logger.info(
        f"CT: Starting Coords for {_corporation.corporation.corporation_name}")

    catagories = [
        23,  # Starbases
        65,  # Structures
        40,  # Sov
        22,  # Deployables
    ]

    # These must be in space
    max_location_id = 35000000
    min_location_id = 30000000

    assets = CorpAsset.objects.filter(
        corporation=_corporation,
        type_name__group__category_id__in=catagories,
        location_id__gte=min_location_id,
        location_id__lte=max_location_id
    )

    logger.info(f"CT: COORDS {assets.count()} Assets found.")

    _req_scopes = ['esi-assets.read_corporation_assets.v1']
    _req_roles = ['Director']
    _token = get_corp_token(
        _corporation.corporation.corporation_id,
        _req_scopes,
        _req_roles
    )

    if not _token or not assets.count():
        logger.info(
            f"CT: COORDS No Tokens or Assets for {_corporation.corporation.corporation_name}")
        return f"CT: COORDS No Tokens or Assets for {_corporation.corporation.corporation_name}"

    _all_ids = assets.values_list("item_id", flat=True)
    locations = []

    locations = []

    for id_chunk in providers.esi_openapi.chunk_ids(_all_ids):
        locations += providers.esi_openapi.client.Assets.PostCorporationsCorporationIdAssetsLocations(
            corporation_id=_corporation.corporation.corporation_id,
            body=id_chunk,
            token=_token
        ).result(
            force_refresh=force_refresh
        )

    logger.info(locations)

    new_coords = {}
    for loc in locations:
        new_coords[loc.item_id] = loc.position

    new_models = []
    for a in assets:
        if a.item_id in new_coords:
            new_models.append(
                AssetCoordiante(
                    item=a,
                    x=new_coords[a.item_id].x,
                    y=new_coords[a.item_id].y,
                    z=new_coords[a.item_id].z
                )
            )
            logger.info(
                f"CT: COORD - {a.type_name.name} {new_coords[a.item_id]}")

    AssetCoordiante.objects.bulk_create(new_models, ignore_conflicts=True)

    return f"CT: Finished Coords for {_corporation.corporation.corporation_name}"


def corp_update_asset_names(corp_id, force_refresh: bool = False):
    """
        Uses PostCorporationCorporationIdAssetsNames
    """

    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("Updating Assets for: {}".format(
        audit_corp.corporation))

    req_scopes = ['esi-assets.read_corporation_assets.v1',
                  'esi-characters.read_corporation_roles.v1']
    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        return "No Tokens"

    expandable_cats = [2, 6, 65]

    asset_list = CorpAsset.objects.filter(
        corporation=audit_corp,
        type_name__group__category_id__in=expandable_cats,
        singleton=True
    ).order_by("pk")

    for subset in chunks(asset_list, 900):
        assets_names = providers.esi_openapi.client.Assets.PostCorporationsCorporationIdAssetsNames(
            corporation_id=corp_id,
            body=list(set([i.item_id for i in subset])),
            token=token
        ).results(
            force_refresh=force_refresh
        )

        id_list = {i.item_id: i.name for i in assets_names}

        for asset in subset:
            if asset.item_id in id_list:
                asset.name = id_list.get(asset.item_id)
                asset.save()

    return f"CT: Finished corp asset names for: {audit_corp.corporation}"


@shared_task(bind=True, base=QueueOnce)
def run_ozone_levels(self, corp_id):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id
    )
    logger.debug("Updating Ozone for: %s" % (_corporation.corporation))

    _structures = Structure.objects.filter(
        type_id=35841,
        corporation=_corporation
    )

    for structure in _structures:
        _quantity = CorpAsset.objects.filter(
            corporation=_corporation,
            location_id=structure.structure_id,
            type_id=16273
        ).aggregate(
            ozone=Sum('quantity')
        )['ozone']
        _used = 0

        try:
            last_ozone = BridgeOzoneLevel.objects.filter(
                station_id=structure.structure_id
            ).order_by('-date')[:1][0].quantity
            delta = last_ozone - _quantity
            _used = (delta if _quantity < last_ozone else 0)
        except Exception:
            pass
        try:
            BridgeOzoneLevel.objects.create(
                station_id=structure.structure_id, quantity=_quantity, used=_used)
        except Exception:
            pass  # dont fail for now
    return f"Finished Ozone for: {_corporation.corporation}"



# Standard Library
from typing import TYPE_CHECKING

# Third Party
from eve_sde.models import ItemType, SolarSystem

# Django
from django.db import models

from .audits import CharacterAudit

if TYPE_CHECKING:
    # Alliance Auth
    from esi.stubs import CharactersCharacterIdMiningGetItem


class CharacterMiningLedger(models.Model):
    id = models.CharField(max_length=50, primary_key=True)
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)
    date = models.DateField()
    type_name = models.ForeignKey(ItemType, on_delete=models.CASCADE)
    system = models.ForeignKey(SolarSystem, on_delete=models.CASCADE)
    quantity = models.IntegerField()

    @staticmethod
    def create_primary_key(character_id, mining_record: "CharactersCharacterIdMiningGetItem"):
        """
            TODO investigate something else...
        """
        return f"{mining_record.date.strftime('%Y%m%d')}-{mining_record.type_id}-{character_id}-{mining_record.solar_system_id}"
